#ifndef _PROforma_H
#define _PROforma_H

/*
	PROforma, PROGS Font & Raster Manager, header file
	definition of all PROforma commands, types and macros

	(c) PROGS, Professional & Graphical Software
	written by Joachim Van der Auwera
*/

/*
	most parameters which are passed to PROforma are fixpoints, as they give
	more accuracy than integers, and more speed than floats or doubles.
	A fixpoint number is actually a long, with an imaginary dot between the two
	words. This convention allows for fast adding and multiplication, without
	loosing too much accuracy
*/

#include <err.h>
#include <io.h>

typedef long pt;
#define pt_one (pt)0x10000
#define pt_half (pt)0x8000
#define pt_quarter (pt)0x4000
#define pt_hundred (pt)0x640000
#define ptmin (pt)0xc0000000
#define ptmax (pt)0x3fffffff
#define ptmagic (pt)36045			/* .55 in fixpoint */
#define ptcigam (pt)29491			/* .45 in fixpoint */

/*
	Most operations can be done directly on fixed-point quantities:
	addition, subtraction, shifting, multiplication or division by
	integer constants; assignment, assignment with zero;
	comparison, comparison against zero.
	Multiplication and division by floats is OK if the result is
	explicitly cast back to fixed.
	Conversion to and from int and float types must be done explicitly.
	Note that if we are casting a fixed to a float in a context where
	only ratios and not actual values are involved, we don't need to take
	the scale factor into account: we can simply cast to float directly.
*/

#define long2pt(x)		((pt)((x)<<16))
#define short2pt(x) 	((pt)(x)<<16)
#define pt2short(x) 	((short)(((unsigned)(x))/65536))
#define pt2rshort(x)	pt2short((x)+pt_half)
#define pt2long(x)		(long)(pt2short(x))
#define pt2rlong(x) 	pt2long((x)+pt_half)
#define pt2int(x)		(int)(pt2short(x))
#define pt2rint(x)		pt2int((x)+pt_half)
#define double2pt(x)	((pt)((x)*(double)65536.0))
#define pt2double(x)	((double)((x)/(65536.0)))

/* Rounding and truncation on fixeds */
#define pt_trunc(x) ((x)&(0xffff0000))
#define pt_round(x) pt_trunc(x+pt_half)
#define pt_ceiling(x) pt_trunc(x+pt_one)
#define pt_fraction(x) ((x)&0xffff)
#define pt_center(x) (pt_trunc(x)+pt_half) /* rounding to the nearest half */

/* there is a special multiplication routine for fixpoint coordinates */
pt fixmul(pt x, pt y);	/* returns x*y */
pt fixdiv(pt x, pt y);	/* returns x/y */
pt fixsqrt(pt x);		/* returns sqrt(x) */
pt fixdist(pt a, pt b); /* returns sqrt( a*a+b*b ) */

/* types of bitmap pictures */
#define PF_PICTURE_QL4SCREEN	1
#define PF_PICTURE_QL8SCREEN	2
#define PF_PICTURE_PIC			3
#define PF_PICTURE_PAINTER4 	4
#define PF_PICTURE_PAINTER8 	5

/* CharPath elements */
#define PF_CHARPATH_MOVETO		1
#define PF_CHARPATH_LINETO		2
#define PF_CHARPATH_CURVETO 	3
#define PF_CHARPATH_END 		5
#define PF_CHARPATH_WIDTH		6

/* methods for drawing a path */
#define PF_PATH_UNDEF		0	/* type of path is undefined */
#define PF_PATH_STROKED 	1	/* path is stroked with thick lines */
#define PF_PATH_FILLED		2	/* path is filled with winding rule */
#define PF_PATH_EOFILLED	3	/* path is filled with even/odd rule */

#define PF_MAXFONTNAME		60
#define PF_MAXFONTFAMILY	60
#define PF_MAXFONTNOTICE	80
#define PF_MAXFONTVERSION	10
#define PF_MAXFONTWEIGHT	20
#define PF_MAXDRIVERNAME	60

#define PF_DRIVER_DUMMY 	-1
#define PF_DRIVER_SCREEN	0
#define PF_DRIVER_DEFAULT	1

#define PF_DISPLAYMODE_BUFFER	0
#define PF_DISPLAYMODE_NORMAL	0
#define PF_DISPLAYMODE_UPDATE	1

#define PF_PATTERNMASK_DITHER		1
#define PF_PATTERNMASK_DOT			2
#define PF_PATTERNMASK_SMALLDOT 	3
#define PF_PATTERNMASK_SQUARE		4
#define PF_PATTERNMASK_SMALLSQUARE	5
#define PF_PATTERNMASK_CROSSH		6
#define PF_PATTERNMASK_CROSSD		7
#define PF_PATTERNMASK_CORNER		8
#define PF_PATTERNMASK_SQUAREH		9

/* access to PROforma thing for dynamic configuration */
#define PF_THING_NAME	"PROforma DLL"
#define PF_THING_CNFG	0x434e4647	/* "CNFG" */

/* macro to convert printer or picture driver names to their id */
#define PF_NAME(x)		((int)x)

typedef struct _Gstate *Gstate;

typedef struct { pt red, green, blue; } ColourRGB;
typedef struct { pt cyan, magenta, yellow, black; } ColourCMYK;

typedef struct { pt xsiz, ysiz, xorg, yorg; } Bbox;
typedef struct { pt xsiz, ysiz, xorg, yorg, xporg, yporg; } PageBbox;
typedef struct { short xsiz, ysiz, xorg, yorg; } PBbox;
typedef struct { short xsiz, ysiz, xorg, yorg;
				 pt xporg, yporg; } PPageBbox;

typedef struct {
	short code;
	char *name;
} CharName;

extern short *PFASCII2UnicodeTable;
extern CharName *PFCharNameTable;

void  LinkPROforma			();

Error PFGstateInit			(int, const char *, const Bbox *, Gstate *, int *);
Error PFGstateRemove		(Gstate);
Error PFGstateSave			(Gstate);
Error PFGstateRestore		(Gstate);

Error PFLineWidth			(Gstate, pt);
Error PFFlatness			(Gstate, pt);

Error PFCTMSave 			(Gstate);
Error PFCTMRestore			(Gstate);
Error PFCTMRestoreKeep		(Gstate);
Error PFCTMReset			(Gstate);
Error PFCTMSet				(Gstate, pt, pt, pt, pt, pt, pt);
#define PFCTMMove(gid, xd, yd)	PFCTMSet(gid, pt_one,0,0,pt_one, xd, yd)
#define PFCTMScale(gid, s)		PFCTMSet(gid, s, 0,0, s, 0,0)
#define PFCTMXScale(gid, xs)	PFCTMSet(gid, xs, 0,0,pt_one,0,0)
#define PFCTMYScale(gid, ys)	PFCTMSet(gid, pt_one,0,0, ys, 0,0)

Error PFPathMethod			(Gstate, int);
Error PFPathClose			(Gstate);
Error PFPathClear			(Gstate);
Error PFPathDraw			(Gstate);
Error PFPathClip			(Gstate);
#define PFPathStroked(g)	PFPathMethod(g,PF_PATH_STROKED)
#define PFPathFilled(g) 	PFPathMethod(g,PF_PATH_FILLED)
#define PFPathEOFilled(g)	PFPathMethod(g,PF_PATH_EOFILLED)

Error PFMoveTo				(Gstate, pt, pt);
Error PFMoveR				(Gstate, pt, pt);
Error PFLineTo				(Gstate, pt, pt);
Error PFLineR				(Gstate, pt, pt);
Error PFCurveTo 			(Gstate, pt, pt, pt, pt, pt, pt);
Error PFCurveR				(Gstate, pt, pt, pt, pt, pt, pt);
Error PFArcTo				(Gstate, pt, pt, pt, pt);
Error PFArcR				(Gstate, pt, pt, pt, pt);
Error PFCircle				(Gstate, pt);
Error PFPie 				(Gstate, pt, pt, pt, pt, pt, char);
Error PFRectangle			(Gstate, pt, pt, pt);

Error PFPageClear			(Gstate);
Error PFPageShow			(Gstate);
Error PFPageScroll			(Gstate, pt, pt);
Error PFPageScale			(Gstate, pt);

Error PFPageBboxReset		(Gstate);
Error PFPageBboxRestore 	(Gstate);
Error PFPageBboxSet 		(Gstate, pt, pt, pt, pt);
Error PFPageBboxGet 		(Gstate, PageBbox *);
Error PFPageBboxFullGet 	(Gstate, Bbox *);
Error PFPageOriginSet		(Gstate, pt, pt);

Error PFClipClear			(Gstate);

Error PFFontScale			(Gstate, pt);
Error PFFontLoad			(Gstate, const char *);
Error PFFontNameGet 		(Gstate, char *);
Error PFFontWeightGet		(Gstate, char *);
Error PFFontFamilyGet		(Gstate, char *);
Error PFFontNoticeGet		(Gstate, char *);
Error PFFontVersionGet		(Gstate, char *);
Error PFFontUnLoad			(Gstate, const char *);
Error PFFontSelect			(Gstate, const char *);
Error PFFontBbox			(Gstate, Bbox *);
Error PFFontCalculate		(Gstate);

Error PFShow				(Gstate, const short *);
Error PFShowKern			(Gstate, const short *);
Error PFShowX				(Gstate, const short *, const pt *);
Error PFShowXY				(Gstate, const short *, const pt *, const pt *);
Error PFShowKernJust		(Gstate, const short *, pt);
Error PFShowJust			(Gstate, const short *, pt);
Error PFShowTrack			(Gstate, const short *, pt);

Error PFWidth				(Gstate, const short *, pt *);
Error PFWidthKern			(Gstate, const short *, pt *);

Error PFFontNext			(Gstate);
Error PFFontNextName		(Gstate, char *);
Error PFFontCount			(Gstate, int *);
Error PFFontCountAll		(Gstate, int *);

Error PFWindowSubRestore	(Gstate);
Error PFWindowSubSet		(Gstate, pt, pt, pt, pt);
Error PFBufferUpdate		(Gstate);
Error PFBufferSave			(Gstate);
Error PFBufferRestore		(Gstate);

Error PFDriverCount 		(Gstate, int *);
Error PFDriverNext			(Gstate, int *, char *);

Error PFCharAvailable		(Gstate, int);
Error PFCharPathInit		(Gstate, int);
Error PFCharPathEl			(Gstate, short *, pt *, pt *, pt *, pt *, pt *, pt *);

Error PFStringShow			(Gstate, const char *);
Error PFStringShowKern		(Gstate, const char *);
Error PFStringShowX 		(Gstate, const char *, const pt *);
Error PFStringShowXY		(Gstate, const char *, const pt *, const pt *);
Error PFStringShowKernJust	(Gstate, const char *, pt);
Error PFStringShowJust		(Gstate, const char *, pt);
Error PFStringShowTrack 	(Gstate, const char *, pt);

Error PFStringWidth 		(Gstate, const char *, pt *);
Error PFStringWidthKern 	(Gstate, const char *, pt *);

Error PFColourGray			(Gstate, pt);
Error PFColourRGB			(Gstate, const ColourRGB *);
Error PFColourCMYK			(Gstate, const ColourCMYK *);
Error PFPaperColourGray 	(Gstate, pt);
Error PFPaperColourRGB		(Gstate, const ColourRGB *);
Error PFPaperColourCMYK 	(Gstate, const ColourCMYK *);
Error PFPatternMask 		(Gstate, short);
Error PFPatternMaskUser 	(Gstate, const char *);

Error PFPrintEscape 		(Gstate, int);
Error PFExtraEOS			(Gstate, int);
Error PFCopies				(Gstate, int);
Error PFCacheUse			(Gstate, int);
Error PFDisplayMode 		(Gstate, int);

Error PFPictureCount		(Gstate, int *);
Error PFPictureNext 		(Gstate, int *, char *);
Error PFPictureRecognize	(Gstate, int, char *);
Error PFPictureRatio		(Gstate, int, char *, int *, int *);
Error PFPictureColourCount	(Gstate, int, char *, int *);
Error PFPictureColourGrayGet(Gstate, int, char *, pt *);
Error PFPictureColourRGBGet (Gstate, int, char *, ColourRGB *);
Error PFPictureColourCMYKGet(Gstate, int, char *, ColourCMYK *);
Error PFPictureDisplay		(Gstate, int, char *, pt, pt);
Error PFPictureDisplayGray	(Gstate, int, char *, pt, pt, pt *);
Error PFPictureDisplayRGB	(Gstate, int, char *, pt, pt, ColourRGB *);
Error PFPictureDisplayCMYK	(Gstate, int, char *, pt, pt, ColourCMYK *);

/* variants for fast pixels accurate access */

Error PFPGstateInit 		(int, const char *, const PBbox *, Gstate *, int *);
Error PFPGstateReAssign 	(Gstate, int, short, short, short, short);
Error PFPGstateReAssign2	(Gstate, int, char *, short, short, short, short);

Error PFPMoveTo 			(Gstate, pt, pt);
Error PFPMoveR				(Gstate, pt, pt);
Error PFPLineTo 			(Gstate, pt, pt);
Error PFPLineR				(Gstate, pt, pt);
Error PFPCurveTo			(Gstate, pt, pt, pt, pt, pt, pt);
Error PFPCurveR 			(Gstate, pt, pt, pt, pt, pt, pt);
Error PFPArcTo				(Gstate, pt, pt, pt, pt);
Error PFPArcR				(Gstate, pt, pt, pt, pt);
Error PFPRectangle			(Gstate, pt, pt);

Error PFPPageScroll 		(Gstate, short, short);

Error PFPPageBboxSet		(Gstate, short, short, short, short);
Error PFPPageBboxGet		(Gstate, PPageBbox *);
Error PFPPageBboxFullGet	(Gstate, PBbox *);
Error PFPPageOriginSet		(Gstate, short, short);

Error PFPShowX				(Gstate, const short *, const short *);
Error PFPShowXY 			(Gstate, const short *, const short *, const short *);
Error PFPStringShowX		(Gstate, const char *, const short *);
Error PFPStringShowXY		(Gstate, const char *, const short *, const short *);

Error PFPWindowMove 		(Gstate, short, short);
Error PFPWindowSubSet		(Gstate, short, short, short, short);

Error PFPShow				(Gstate, const short *);
Error PFPWidth				(Gstate, const short *, short *);
Error PFPStringShow 		(Gstate, const char *);
Error PFPStringWidth		(Gstate, const char *, short *);
Error PFPFontBbox			(Gstate, short, PBbox *);

Error PFPWindowLink 		(Gstate, Channel, short, short, short, short, short, short);

#endif _PROforma_H
